-- ----------------------------------------------------------------------------
-- Search for a Connected by TCP gateway via SSDP
-- ----------------------------------------------------------------------------
function tcplighting_gateway_ssdp_search ()

    -- Create an SSDP request - note that the Connected By TCP gateways
    -- do not appear to respond to ssdp:all requests, so we'll directly
    -- target the Greenwave Reality gateways that TCPi uses
    -- See: http://upnp.org/specs/arch/UPnP-arch-DeviceArchitecture-v1.1.pdf
    local ssdp_headers =
        {
        [1] = "M-SEARCH * HTTP/1.1\r\n",
        [2] = "HOST: 239.255.255.250:1900\r\n",
        [3] = "MAN: \"ssdp:discover\"\r\n",
        [4] = "MX: 3\r\n",
        [5] = "ST: urn:greenwavereality-com:service:gop:1\r\n\r\n"
        }
    local ssdp_request = table.concat (ssdp_headers )

    -- Get a UDP socket
    local http = require("socket.http")
    local udp = socket.udp()

     -- Send out the SSDP request.  Don't wait forever, they normally respond quickly or not at all.
    udp:sendto (ssdp_request, "239.255.255.250", 1900)
    udp:settimeout (3)

    -- Loop over response packets until we get and IP or there is no more data
    local gateway_ip   = nil
	local ssdp_uuid    = nil

    repeat
        local received_datagram = udp:receive()
        if received_datagram then
           -- pull the gateway IP out of the LOCATION header in the response
           gateway_ip = string.match (received_datagram, "LOCATION:[%s+]http://([%d%.]+)")
           -- pull the gateway UID out of the USN header
           ssdp_uuid  = string.match (received_datagram, "USN:[%s+]uuid:(%d+)")
        else
           break
        end
    until (gateway_ip)

    -- return the IP address & UUID of the discovered gateway (or nil if we are SOL)
    return gateway_ip, ssdp_uuid
end

-- ----------------------------------------------------------------------------
-- Pair with a Connected by TCP gateway
--
-- Note: User must put gateway into "pairing" mode by pressing 
--       green sync button (green on the model I have anyway)
-- ----------------------------------------------------------------------------
function tcplighting_gateway_pair (gateway)

    -- Bail if we don't a gateway
    if (not gateway) then
       gateway.error_text = "No gateway object provided, pairing failed"
       return gateway
    end

    -- Bail if we don't a gateway IP
    if (not gateway.gateway_ip) then
       gateway.error_text = "No gateway IP address specified, pairing failed"
       return gateway
    end

    -- Bail if we  don't have a gateway UUID
    if (not gateway.gateway_uuid) then
       gateway.error_text = "No gateway UUID address derived, pairing failed"
       return 
    end

    local http  = require("socket.http")
    local https = require("ssl.https")

    http.TIMEOUT = 5
    
    local result, status  = https.request("https://" .. gateway.gateway_ip
                                          .. "/gwr/gop.php",
                                         "cmd=GWRLogin&data="
                                          .. "<gip>"
                                          .. "<version>1</version>"
                                          .. "<email>" .. gateway.gateway_uuid .. "</email>"
                                          .. "<password>" .. gateway.gateway_uuid .. "</password>"
                                          .. "</gip>"
                                          .. "&fmt=xml")
    
    local rc_value    = nil
    if (status == 200 and result) then
       -- The result is a small bit of XML that can easily be parsed through string manipulation
       rc_value    = extract_xml_element_value (result, "rc")
       gateway.gateway_token = extract_xml_element_value (result, "token")
       gateway.error_text = ""
    end

    if (not gateway.gateway_token) then
       gateway.error_text = "No gateway token extracted, pairing failed"
       if status   then gateway.error_text = gateway.error_text .. " - HTTPS status=" .. status  end
       if rc_value then gateway.error_text = gateway.error_text .. " - Gateway RC=" .. rc_value  end
    end

    return 

end


-- ----------------------------------------------------------------------------
-- Implements urn:upnp-org:serviceId:SwitchPower1.SetTarget
-- ----------------------------------------------------------------------------
function tcplighting_set_target (gateway, did, target_value)

    local http  = require("socket.http")
    local https = require("ssl.https")

    http.TIMEOUT = gateway.HTTP_TIMEOUT

    return https.request("https://" .. gateway.devices[did].gateway_ip
                        .. "/gwr/gop.php",
                        "cmd=DeviceSendCommand&data="
                        .. "<gip>"
                        .. "<version>1</version>"
                        .. "<token>" .. gateway.gateway_token .. "</token>"
                        .. "<did>" .. did .. "</did>"
                        .. "<value>" .. target_value .. "</value>"
                        .. "</gip>"
                        .. "&fmt=xml")
end

-- ----------------------------------------------------------------------------
-- Implements urn:upnp-org:serviceId:Dimming1.SetLoadLevelTarget
-- ----------------------------------------------------------------------------
function tcplighting_set_load_level_target (gateway, did, load_percent)

    local http  = require("socket.http")
    local https = require("ssl.https")

    http.TIMEOUT = gateway.HTTP_TIMEOUT

    -- make sure light is on/off to match request
    if load_percent > 0 then
        tcplighting_set_target (gateway, did, 1)
    else
        tcplighting_set_target (gateway, did, 0)
    end

    -- change the dimming level
    return https.request("https://" .. gateway.devices[did].gateway_ip
                        .. "/gwr/gop.php",
                        "cmd=DeviceSendCommand&data="
                        .. "<gip>"
                        .. "<version>1</version>"
                        .. "<token>" .. gateway.gateway_token .. "</token>"
                        .. "<did>" .. did .. "</did>"
                        .. "<value>" .. load_percent .. "</value>"
                        .. "<type>level</type>"
                        .. "</gip>"
                        .. "&fmt=xml")
end


-- ----------------------------------------------------------------------------
-- Create a gateway object
-- ----------------------------------------------------------------------------
function tcplighting_gateway_create (gateway_ip)

    -- create an empty gateway object
    local gateway = {}
    gateway.authenticated  = false
    gateway.initialized    = false
    gateway.autodiscovered = false
    gateway.error_text     = nil    -- error text for last failed operation
    gateway.gateway_ip     = gateway_ip
    gateway.gateway_uuid   = nil
    gateway.gateway_token  = nil
    gateway.ssdp_uuid      = nil
    gateway.devices        = {}    -- a cache of device info indexed by unique did
    gateway.HTTP_TIMEOUT   = 5     -- 5 seconds should be fine on a local network

    return gateway

end


-- ----------------------------------------------------------------------------
-- Loads the gateway configuration
-- ----------------------------------------------------------------------------
function tcplighting_gateway_init (gateway_ip)

    local MAX_SSDP_RETRIES = 5

    -- create a gateway object
    local gateway = tcplighting_gateway_create (gateway_ip)

	-- Attempt auto-discovery via SSDP if IP not specified (not likely to work with new "secure" gateways)
    if (gateway_ip == nil) or (string.len(gateway_ip) == 0) then
        -- since UDP doesn't provide a reliable transport, try the SSDP request few times if necessary
        for i = 1, MAX_SSDP_RETRIES do
           gateway.gateway_ip, gateway.ssdp_uuid = tcplighting_gateway_ssdp_search ()
           if gateway.gateway_ip then
		      gateway.autodiscovered = true
		      break
		   end
        end
	else
	    -- used specified IP
	    gateway.gateway_ip = gateway_ip
	end

    -- Bail if we still don't have an IP
    if not gateway.gateway_ip then
       gateway.error_text = "No gateway IP address specified, autodiscovery failed"
       return gateway
    end

    return gateway

end


-- ----------------------------------------------------------------------------
-- Get current values from the gateway
-- ----------------------------------------------------------------------------
function tcplighting_gateway_poll (gateway)

    -- Bail if we don't a gateway IP
    if (not gateway.gateway_ip) then
       gateway.error_text = "No gateway IP address specified, polling failed"
       return gateway
    end

    -- Get the gateway's configuration data
    local result, status = tcplighting_get_gateway_config (gateway)

    -- Check the HTTP status
    if status ~= 200 then
       gateway.error_text = "HTTP gateway query request to " .. gateway.gateway_ip .. " failed"
       if status then gateway.error_text = gateway.error_text .. " - " .. status     end
       if result then gateway.error_text = gateway.error_text .. " - " .. result     end
       return
    end

    -- Make sure something came back
    if not result then
       gateway.error_text = "HTTP gateway query request to " .. gateway.gateway_ip .. " did not return any data"
       return
    end

    -- Only parse if the HTTP request was successful
    tcplighting_parse_gateway_config(gateway, result)
    return

end


-- ----------------------------------------------------------------------------
-- Queries a Connected by TCP Gateway to determine which lights are available
-- ----------------------------------------------------------------------------
function tcplighting_get_gateway_config (gateway)

    local http  = require("socket.http")
    local https = require("ssl.https")


    http.TIMEOUT = gateway.HTTP_TIMEOUT

    return https.request("https://" .. gateway.gateway_ip
                        .. "/gwr/gop.php",
                        "cmd=GWRBatch&data="
                        .. "<gwrcmds>"
                        .. "<gwrcmd>"
                        .. "<gcmd>RoomGetCarousel</gcmd>"
                        .. "<gdata>"
                        .. "<gip>"
                        .. "<version>1</version>"
                        .. "<token>" .. gateway.gateway_token .. "</token>"
                        .. "<fields>name,image,imageurl,control,power,product,class,realtype,status</fields>"
                        .. "</gip>"
                        .. "</gdata>"
                        .. "</gwrcmd>"
                        .. "</gwrcmds>"
                        .. "&fmt=xml")
end


-- ----------------------------------------------------------------------------
-- Parse the gateway configuration
-- ----------------------------------------------------------------------------
function tcplighting_parse_gateway_config (gateway, config_xml)

    local lxp = require("lxp")

    -- Local collection devices - will be added to master table after parsing
    local new_devices = {}

    -- enumerate the XML elements that we want to load into our tables
    local room_elements_of_interest = "|rid|name|desc|"
    local device_elements_of_interest = "|did|name|desc|state|level|offline|type|rangemin|rangemax|power|poweravg|energy|manufacturer|prodbrand|prodmodel|prodtype|capability|"

    -- used to keep track of the current context during the parsing process
    local current_room_object = nil
    local current_device_object = nil
    local current_element_name = nil

    -- XML parsing handler for the start of an element
    local function start_element_handler (parser, name)
        current_element_name = name
        if name == "room" then
            current_room_object = {}
        elseif name == "device" then
            current_device_object = {}
            current_device_object["gateway_ip"] = gateway.gateway_ip
            current_device_object["room"] = current_room_object
        end
    end

    -- XML parsing handler for the end of an element
    local function end_element_handler (parser, name)
        current_element_name = name
        if name == "room" then
            current_room_object = nil
        elseif name == "device" then
             new_devices[#new_devices+1] = current_device_object
            current_device_object = nil
        end
    end

    -- XML parsing handler for the text of an element
    local function character_handler (parser, value)
        if current_device_object and string.find(device_elements_of_interest,"|"..current_element_name.."|") then
            current_device_object[current_element_name] = value
        elseif current_room_object and (not current_device_object) and string.find(room_elements_of_interest,"|"..current_element_name.."|") then
            current_room_object[current_element_name] = value
        end
    end

    -- Create a parser
    local callbacks = {StartElement = start_element_handler, EndElement = end_element_handler, CharacterData = character_handler}
    local parser = lxp.new(callbacks)

    -- Parse the XML
    parser:parse(config_xml)
    parser:parse() -- finishes the document
    parser:close()

    -- Save the devices, uniquely indexed by did
    if #new_devices > 0 then
        for i,device in ipairs(new_devices) do
            gateway.devices[device.did] = device
            -- tweak for interpretation of load level for dimmers - always set to zero if device is off
            if device.state == "0" then
               device.level = "0"
            end
        end
        gateway.initialized = true
        gateway.error_text = nil
    else
        gateway.error_text = "No devices found at gateway " .. gateway.gateway_ip .. " while parsing " .. config_xml
    end

    return

end


-- ----------------------------------------------------------------------------
-- Extracts the contents of the first occurrence of the given element 
-- Note: element_name should not contain angle brackets (i.e. "xxx", not "<xxx>")
-- ----------------------------------------------------------------------------
function extract_xml_element_value (xml_string, element_name)
    local element_value = nil
    if (xml_string and element_name) then
       local element_start_pos = string.find (xml_string, "<" .. element_name .. ">")
       local element_end_pos   = string.find (xml_string, "</" .. element_name .. ">")
       if (element_start_pos and element_end_pos) then
           element_value = string.sub (xml_string, element_start_pos+string.len(element_name)+2, element_end_pos-1)
       end
    end
    return element_value
end


-- ----------------------------------------------------------------------------
-- Helper functions for testing - call devices by name/room
-- ----------------------------------------------------------------------------
function tcplighting_get_did (gateway, name, roomname)
   local did = nil
   for key,device in pairs(gateway.devices) do
      if device.name == name and device.room.name == roomname then
         did = device.did
      end
   end
   return did
end

function tcplighting_set_target_by_name  (gateway, name, roomname, target_value)
    return tcplighting_set_target(gateway, tcplighting_get_did (gateway, name, roomname), target_value)
end

function tcplighting_set_load_level_target_by_name (gateway, name, roomname, load_percent)
    return tcplighting_set_load_level_target(gateway, tcplighting_get_did (gateway, name, roomname), load_percent)
end
